const CHARS_PER_STROKE = 5;
const charTiming = new Map();
charTiming.set(undefined, .025);
charTiming.set(",", .2);
charTiming.set(":", .2);
charTiming.set(";", .2);
charTiming.set(".", .5);
charTiming.set("?", .5);
charTiming.set("!", .5);
charTiming.set("_", .4);

// charTiming.set(" ", .3);
// charTiming.set(",", .4);
// charTiming.set(";", .5);
// charTiming.set("_", .4);
// charTiming.set(undefined, 0);

const AUDIO = new Map();
const IMAGES = new Map();
const VIDEOS = new Map();
const FONTS = new Map();

const EVENTS = new Map();
const SPEAKERS = new Map();

const SIGNALS = new EventTarget();
function ADVANCE() {
    SIGNALS.dispatchEvent(new Event("advance"));
    REFOCUS();
}

var MUSIC = undefined;
/** @type {HTMLElement} */
var FOCUS = undefined;

function SET_FOCUS(element) {
    FOCUS = element;
    REFOCUS();
}

function DEL_FOCUS(element) {
    if (FOCUS === element) {
        FOCUS?.blur();
        FOCUS = undefined;
    }
}

function REFOCUS() {
    if (document.activeElement != document.body) return;
    FOCUS?.focus();
}

const DELAY = (seconds) => sleep(seconds * 1000);

async function WAIT_AUDIO_ALLOWED() {
    while (Howler.state === "suspended") await DELAY(.1);
}

function PLAY_SOUND(id) {
    AUDIO.get(id)?.play();
}

function PLAY_MUSIC(id) {
    const next = AUDIO.get(id);
    if (next !== MUSIC) MUSIC?.fade(MUSIC._volume, 0, 1000);
    
    MUSIC = next;
    MUSIC?.stop();
    MUSIC?.volume(MUSIC._initial_volume ?? 1);
    MUSIC?.play();
} 

function STOP_MUSIC() {
    PLAY_MUSIC();
}

function SET_EVENT(id, func) {
    EVENTS.set(id, func);
}

async function RUN_EVENT(id) {
    const event = EVENTS.get(id);
    
    if (!event) {
        console.log(`NO EVENT "${id}"`);
    } else {
        try {
            await event();
        } catch (e) {
            console.log(`EVENT "${id}" FAILED`);
            throw e;
        }
    }
}

function SET_SPEAKER(id, func) {
    SPEAKERS.set(id, func);
}

async function RUN_SPEAKER(id, ...texts) {
    const speaker = SPEAKERS.get(id);
    
    if (!speaker) {
        console.log(`NO SPEAKER "${id}" for ${texts}`);
    } else {
        await speaker(...texts);
    }
}

function promiseProgress(promises, callback) {
    if (callback === undefined) return;
    let complete = 0;
    callback(0);
    promises.forEach((promise) => promise.then(() => { complete += 1; callback(complete / promises.length); }));
    return Promise.all(promises);
}

async function PRELOAD_FONTS(callback) {
    return promiseProgress([...FONTS.values()].map((font) => font.load()), callback);
}

async function PRELOAD_IMAGES(callback) {
    const images = [...IMAGES.values()].map((src) => loadImage(src));
    const videos = [...VIDEOS.values()].map((src) => loadVideo(src));
    return promiseProgress([...images, ...videos], callback);
}

async function PRELOAD_MUSIC(callback) {
    const music = [...AUDIO.values()].map((howl) => loadAudio(howl._src));
    return promiseProgress(music, callback);
}

class Panel extends EventTarget {
    constructor(...elements) {
        super();
        this.root = html("div", { class: "panel" }, ...elements);
    }

    attach(parent) {
        parent.append(this.root);
    }

    show() {
        this.root.hidden = false;
    }

    hide() {
        this.root.hidden = true;
    }

    async in() {
        this.root.hidden = false;
        this.dispatchEvent(new Event("pre-in"));
        await animateElement(this.root, "grow-in", .25);
    }

    async out() {
        await animateElement(this.root, "grow-out", .25);
        this.root.hidden = true;
    }

    remove() {
        this.root.remove();
    }
}

class TextAnimation {
    constructor(text, voice, wait=true) {
        this.voice = voice;
        this.cursor = html("div", { class: "cursor" }, "_");

        this.glyphs = Array.from(text).map((char) => ({
            char,
            element: html("span", { class: "char-hidden" }, char),
            delay: charTiming.get(char) ?? charTiming.get(undefined),
            time: 0,
            pause: false,
        }));

        for (let i = this.glyphs.length - 1; i > 0; --i) {
            if (this.glyphs[i].char === "@") {
                this.glyphs.splice(i, 1);
                this.glyphs[i-1].pause = true;
                this.glyphs[i-1].delay = 0;
            } else if (this.glyphs[i].char === "_") {
                this.glyphs.splice(i, 1);
                this.glyphs[i-1].delay = charTiming.get("_") ?? charTiming.get(undefined);
            }
        }

        if (wait) {
            this.glyphs[this.glyphs.length - 1].pause = true;
        }

        let time = 0;
        for (const glyph of this.glyphs) {
            glyph.time = time;
            time += glyph.delay;
        }

        this.time = 0;
        this.paused = false;
        this.done = false;

        this.revealCount = 0;
    }

    update(dt) {
        if (this.done || this.paused) return;

        this.cursor.remove();

        const prev = this.time;
        const next = prev + dt;
        this.time = next;

        let last = undefined;
        let stroke = false;
    
        for (const glyph of this.glyphs) {
            if (prev <= glyph.time && glyph.time < next) {
                glyph.element.classList.remove("char-hidden");
                glyph.element.classList.toggle("char-reveal", true);
                
                if (glyph.char !== " ") {
                    stroke = true;
                    this.revealCount += 1;
                }

                last = glyph;

                if (glyph.pause) {
                    this.paused = true;
                    this.revealCount = 0;
                    break;
                }
            }
        }

        stroke = stroke && this.revealCount % CHARS_PER_STROKE === 0;
        if (stroke) this.voice.stroke();

        this.done = this.glyphs[this.glyphs.length - 1].time < next && !this.paused;

        const halted = this.done || this.paused;

        if (halted && last !== undefined) {
            last?.element.append(this.cursor);
            // sleep(250).then(() => this.voice.finish());
            this.voice.finish();
        }
    }

    proceed() {
        this.paused = false;
    }

    hide() {
        for (const glyph of this.glyphs) {
            glyph.element.classList.toggle("char-hidden", true);
        }
    }

    remove() {
        this.glyphs.forEach((glyph) => glyph.element.remove());
    }

    finish() {
        this.cursor.remove();
    }
}

class Voice {
    constructor(stroke, finish, clear) {
        this.strokeSound = stroke;
        this.finishSound = finish;
        this.clearSound = clear;
    }

    stroke() {
        AUDIO.get(this.strokeSound)?.play();
    }

    finish() {
        AUDIO.get(this.finishSound)?.play();
    }

    clear() {
        AUDIO.get(this.clearSound)?.play();
    }
}

const VOICES = new Map();


function testAnim(anim) {
    return new Promise((resolve) => {
        let prev;

        function update(timestamp) {
            const dt = Math.min((timestamp - (prev ?? timestamp)) / 1000, 1/15);
            prev = timestamp;
            anim.update(dt);

            if (!anim.done) requestAnimationFrame(update);
            else resolve();
        }

        requestAnimationFrame(update);
    });
}

const singleLineDialoguePattern = /^\s*\[([^\s]+)\]\s+(.+)$/;

function parseTest(transcript) {
    const lines = transcript.split("\n");
    const chunks = [];
    let current = { id: "", commands: [] };
    chunks.push(current);

    while (lines.length > 0) {
        const prev = lines.length;

        const pattern = /^\s*\#\s*(.*)$/;
        if (pattern.test(lines[0])) {
            const line = lines.shift();
            const [, id] = line.match(pattern);
            current = { id: id.trim(), commands: [] };
            chunks.push(current);
        }

        checkEvent(lines, current.commands);
        checkJavascript(lines, current.commands);
        checkSingleLine(lines, current.commands);
        checkMultiLine(lines, current.commands);

        if (lines.length === prev) lines.shift();
    }

    return chunks;
}

function checkEvent(lines, output) {
    const pattern = /^\s*([\-\>])\s*(.*)$/;

    if (pattern.test(lines[0])) {
        const line = lines.shift();
        const [, symbol, id] = line.match(pattern);
        output.push({ type: "event", id: id.trim(), wait: symbol === ">" });
    }
}

function checkJavascript(lines, output) {
    const pattern = /^\s*\$\s+(.*)$/;

    if (pattern.test(lines[0])) {
        const line = lines.shift();
        const [, text] = line.match(pattern);
        output.push({ type: "javascript", text: text.trim() });
    }
}

function checkSingleLine(lines, output) {
    if (singleLineDialoguePattern.test(lines[0])) {
        const line = lines.shift();
        const [, id, text] = line.match(singleLineDialoguePattern);
        const texts = text.split(/\/\//).map((t) => t.trim());
        output.push({ type: "dialogue", id, texts });
    }
}

function checkMultiLine(lines, output) {
    const startPattern = /^\s*\[([^\]]+)\]\s*$/;
    const midPattern = /^\s*\/{2}\s*$/;
    const endPattern = /^\s*\-{3}\s*$/;

    if (startPattern.test(lines[0])) {
        const [, id] = lines.shift().match(startPattern);
        const texts = [];
        let text = "";

        while (lines.length > 0) {
            if (midPattern.test(lines[0])) {
                texts.push(text);
                text = "";
                lines.shift();
            } else if (endPattern.test(lines[0])) {
                texts.push(text);
                text = "";        
                lines.shift();
                break;
            } else {
                if (text.length > 0) text += " ";
                text += lines[0].trim();
                lines.shift();
            }
        }

        output.push({ type: "dialogue", id, texts });
    }
}
