document.addEventListener("DOMContentLoaded", setup);

async function setup() {
    const safe = document.getElementById("safe-container");

    // window scaling
    function update() {
        scaleElementToParent(safe);
        requestAnimationFrame(update);
    }
    update();
}

/**
 * @param {string} selector
 */
function grabDataTemplate(selector) {
    const template = /** @type {HTMLTemplateElement} */ (ONE(selector));
    
    /**
     * @param {string} selector
     * @param {(function(HTMLElement):void} func
     */
    function processElements(selector, func) {
        ALL(selector, template.content).forEach((element) => func(element));
    }

    return processElements;
}

function indexData(id) {
    const processElements = grabDataTemplate(`[id="${id}"]`);

    processElements("audio", (element) => {
        const id = element.getAttribute("id");
        const audio = new Howl({ 
            src: element.getAttribute("src"), 
            volume: parseFloat(element.getAttribute("volume") ?? "1"),
            loop: element.hasAttribute("loop"),
            html5: element.dataset.stream !== undefined,
        });
        AUDIO.set(id, audio);
        audio._initial_volume = audio._volume;
        SET_EVENT(`${id}/loop`, () => PLAY_MUSIC(id));
        SET_EVENT(`${id}/play`, () => PLAY_SOUND(id));
        SET_EVENT(`${id}/play-wait`, () => PLAY_SOUND(id));
    });

    processElements("tape-image", (element) => {
        IMAGES.set(element.id, element.dataset.src);
    });

    processElements("tape-video", (element) => {
        VIDEOS.set(element.id, element.dataset.src);
    });

    processElements("tape-font", (data) => {
        const face = new FontFace(data.dataset.family, `url(${data.dataset.src})`);
        document.fonts.add(face);
        FONTS.set(data.id, face);
    });
}

class Panel2 {
    constructor(root) {
        this.root = root;

        this.visible = false;
    }

    show() {
        this.visible = true;
        animateElement(this.root, "grow-in", 0);
    }

    hide() {
        this.visible = false;
        animateElement(this.root, "grow-out", 0);
    }

    async in() {
        if (this.visible) return;
        this.visible = true;
        await animateElement(this.root, "grow-in", .25);
    }

    async out() {
        if (!this.visible) return;
        this.visible = false;
        await animateElement(this.root, "grow-out", .25);
    }

    remove() {
        this.root.remove();
    }
}

async function test() {
    window.addEventListener("click", () => ADVANCE());
    window.addEventListener("keypress", () => ADVANCE());

    indexData("data/preload");

    const safe = document.getElementById("safe-container");
    const dialogueTemplate = document.querySelector("#panels-dialogue");

    /** @type {Map<string, Panel>} */
    const ACTIVE_PANELS = new Map(); 

    SET_EVENT("null", () => {});
    SET_EVENT(`music/stop`, () => PLAY_MUSIC());
    SET_EVENT("hide-all-panels", () => {
        const outs = [...ACTIVE_PANELS.values()].map((panel) => panel.out());
        ACTIVE_PANELS.clear();
        return Promise.all(outs);
    });

    const processElements = grabDataTemplate("[id='scene']");

    processElements("script", (script) => {
        const func = new AsyncFunction("", script.textContent);
        SET_EVENT(script.getAttribute("id"), func);
    });

    processElements("tape-delay", (element) => {
        SET_EVENT(element.id, () => DELAY(cssTimeToMilliseconds(element.dataset.duration) / 1000));
    });

    processElements("tape-background", (data) => {
        SET_EVENT(data.id, async () => {
            const img = IMAGES.get(data.dataset.image);
            const vid = VIDEOS.get(data.dataset.image);
            const media = vid ? html("video", { src: vid, muted: true, autoplay: true, loop: true }) : html("img", { src: img });

            const root = document.querySelector("#background-video");
            // await animateElement(root, "grow-out", .75);
            // root.replaceChildren(media);
            // await animateElement(root, "grow-in", .75);

            const prev = root.querySelector("div");
            media.style.zIndex = "10";
            root.append(media);
            await animateElement(media, "fade-in", .75);
            prev?.remove();
            media.style.zIndex = "0";
        });

        SET_EVENT(`${data.id}/jump`, async () => {
            const img = IMAGES.get(data.dataset.image);
            const vid = VIDEOS.get(data.dataset.image);
            const media = vid ? html("video", { src: vid, muted: true, autoplay: true, loop: true }) : html("img", { src: img });

            const root = document.querySelector("#background-video");
            root.replaceChildren(media);
        });
    });

    function createMediaElement(id) {
        const img = IMAGES.get(id);
        const vid = VIDEOS.get(id);
        
        const media = vid 
            ? html("video", { src: vid, muted: true, autoplay: true, loop: true }) 
            : html("img", { src: img });

        return media;
    }

    processElements("tape-voice", (voice) => VOICES.set(voice.id, new Voice(voice.dataset.stroke, voice.dataset.wait, voice.dataset.clear)));

    function makeChoiceSpeaker(parent) {
        return async (choice, ..._) => {
            const pattern = /\s*([^>]+)\s+>\s+([\w\-\/]+)/gm;
            const matches = [...choice.matchAll(pattern)];
            const options = matches.map(([, label, value]) => ({ label, value }));

            let root = undefined;
    
            const result = await new Promise((resolve) => {
                root = makeChoice(options, resolve, { change: "choice/key-stroke", submit: "choice/key-wait" });
                parent.append(root);
                PLAY_SOUND("choice/key-stroke");
                animateElement(root, "grow-in", .25);
                SET_FOCUS(root.querySelector("input:not([disabled])"));
            });

            DEL_FOCUS(root.querySelector("input"));
    
            await animateElement(root, "grow-out", .25);
            root.remove();
            return RUN_EVENT(result);
        };
    }

    async function runCellSpeaker(parent, options, ...texts) {
        const clone = dialogueTemplate.content.cloneNode(true);

        const portrait = clone.querySelector("#portrait");
        const content = clone.querySelector("#content");
        const elements = [...clone.children];
        parent.append(...elements);

        const anims = texts.map((text) => new TextAnimation(text, VOICES.get(options.voice)));
        const pages = anims.map((anim) => ({ anim, root: html("div", { class: "dialogue-page", style: "opacity: 0" }, ...anim.glyphs.map((g) => g.element))}));
        
        content.classList.add("dialogue-pages");
        content.replaceChildren(...pages.map((p) => p.root));
        
        if (options.portrait) {
            portrait.append(createMediaElement(options.portrait));
            if (options.caption) {
                portrait.append(html("div", { class: "info-text info-text-bottom" }, options.caption));
            }
        } else {
            portrait.remove();
        }

        animateElement(portrait, "grow-in", .25);
        await animateElement(content, "grow-in", .25);

        let prev;

        for (const page of pages) {
            if (prev) prev.root.style.opacity = "0";
            page.root.style.opacity = "1";

            const proceed = () => page.anim.proceed();

            SIGNALS.addEventListener("advance", proceed);
            
            await testAnim(page.anim);
            page.anim.hide();
            prev = page;

            SIGNALS.removeEventListener("advance", proceed);
        }

        animateElement(content, "grow-out", .25);
        await animateElement(portrait, "grow-out", .25);

        elements.forEach((element) => element.remove());
    }

    const transcripts = new Map();

    processElements("tape-transcript", (element) => {
        const chunks = parseTest(element.innerText);
        const all = [];
        let prev;

        chunks.forEach(({ id, commands }, i) => {
            id = `${element.id}/${id}`;
            transcripts.set(id, commands);
            all.push(...commands);

            if (prev) prev.push({ type: "event", id });
            prev = commands;
        });
        transcripts.set(element.id, all);
    });

    const commands = [];
    transcripts.forEach((transcript, id) => {
        SET_EVENT(id, () => {
            commands.length = 0;
            commands.push(...transcript);
        });
    });

    function makeChoiceOption({ label, value }) {
        const input = html("input", { type: "radio", name: "choice", value });
        input.disabled = value === "disabled";
        const div = html("div", { class: "choice-label" }, label);
        const root = html("label", {}, input, div);
        return root;
    }

    function makeChoice(options, resolve, sounds) {
        const opts = options.map(makeChoiceOption);
        const submit = html("input", { type: "submit", hidden: "true" });
        const form = html("form", { class: "choice panel", action: "javascript:void(0);" }, ...opts, submit);

        let selected = undefined;
        opts.forEach((element, i) => {
            const input = element.querySelector("input");

            if (!selected && !input.disabled) {
                input.checked = true;
                selected = input;
            }
            
            input?.addEventListener("input", () => {
                if (input.checked) PLAY_SOUND(sounds.change);
            });
        });
        SET_FOCUS(selected);

        const onSubmit = async () => {
            form.removeEventListener("submit", onSubmit);
            DEL_FOCUS(opts[0]);

            PLAY_SOUND(sounds.submit);

            const data = new FormData(form);
            resolve(data.get("choice"));
        };

        opts.forEach((element) => element.addEventListener("dblclick", () => onSubmit()));
        form.addEventListener("submit", onSubmit);

        return form;
    }

    const alignLookup = {
        "T": "start", "M": "center", "B": "end",
        "L": "left", "C": "center", "R": "right", 
        "*": "stretch",
    }

    function makeGrid(id, cols, rows, rigid) {
        const grid = html("div", { id });
        const min = rigid ? 0 : "auto";
        Object.assign(grid.style, {
            "position": "absolute",
            "top": "0", "right": "0", "bottom": "0", "left": "0",
            "gap": "8px",

            "margin": "12px",
            "margin-top": "16px",

            "display": "grid",
            "grid-template-columns": `repeat(${cols}, minmax(${min}, 1fr))`,
            "grid-template-rows": `repeat(${rows}, minmax(${min}, 1fr))`,

            "place-content": "center",
        });

        return grid;
    }

    function makeBounds({ x0 = 1, y0 = 1, x1 = -1, y1 = -1, halign = "center", valign = "center", zIndex = 0 } = {}) {
        const bounds = html("div");
        Object.assign(bounds.style, {
            "display": "flex",
            "z-index": zIndex.toString(),
            "grid-area": `${y0} / ${x0} / ${y1} / ${x1}`,
            "justify-content": halign,
            "align-items": valign,
        });

        return bounds;
    }
    
    const coordFullPattern = /^(\w)(\w)[\-\:\,](\w)(\w)$/;
    const coordHalfPattern = /^(\w)(\w)$/;

    /**
     * @param {string} text
     */
    function parseCellRange(text) {
        let [x0, y0] = [ 1,  1];
        let [x1, y1] = [-2, -2];

        if (coordFullPattern.test(text)) {
            [, x0, y0, x1, y1] = text.match(coordFullPattern).map(boundCoord);
        } else if (coordHalfPattern.test(text)) {
            [, x0, y0] = text.match(coordHalfPattern).map(boundCoord);
            [x1, y1] = [x0, y0];
        }

        x1 += 1;
        y1 += 1;

        return { x0, y0, x1, y1 };
    }

    const letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    function boundCoord(char) {
        const letter = letters.indexOf(char) + 1;
        return letter > 0 ? letter : parseInt(char);
    }

    const CELLS = new Map();

    processElements("tape-grid", (gridData) => {
        const cols = parseInt((gridData.dataset.cols ?? "3"), 10);
        const rows = parseInt((gridData.dataset.rows ?? "3"), 10);
        const rigid = gridData.dataset.rigid !== undefined;

        const grid = makeGrid(gridData.id, cols, rows, rigid);
        const panelDatas = ALL("tape-panel", gridData);
        const speakerDatas = ALL("tape-speaker", gridData);
        const logDatas = ALL("tape-log", gridData);

        copyStyle(grid, gridData);

        SET_EVENT(`${gridData.id}/dim`, () => animateElement(grid, "dim-out", .25));

        const cells = [];

        SET_EVENT(`${gridData.id}/all/in`, () => {
            return Promise.all(cells.map((id) => RUN_EVENT(`${id}/in`)));
        });

        SET_EVENT(`${gridData.id}/all/out`, () => {
            return Promise.all(cells.map((id) => RUN_EVENT(`${id}/out`)));
        });
        
        [...gridData.children].forEach((e, i) => e._childIndex = i);

        function makePanel(panelData) {
            const range = parseCellRange(panelData.dataset.bounds);
            const align = panelData.dataset.align ?? "CC";
            const halign = alignLookup[align[1]];
            const valign = alignLookup[align[0]];

            const bounds = makeBounds({ ...range, halign, valign });            
            grid.appendChild(bounds);
            bounds._childIndex = panelData._childIndex;

            return bounds;
        }

        panelDatas.forEach((panelData) => {
            const bounds = makePanel(panelData);

            const index = [...panelData.parentElement.children].indexOf(panelData);
            const indexId = `${grid.id}/${index+1}`;

            const cell = html("div", { style: "position: relative;" });
            copyStyle(cell, panelData);
            bounds.append(cell);

            const cellPanel = new Panel2(cell);
            let choiceParent;
 
            if (panelData.dataset.image !== undefined) {
                cell.classList.add("panel");
                cell.append(createMediaElement(panelData.dataset.image));
                // cell.append(html("div", { class: "glass-border" }));

                choiceParent = html("div", { style: "position: absolute; top: 0; right: 0; bottom: 0; left: 0; display: flex; place-items: center; place-content: center" });
                cell.append(choiceParent);
                cells.push(indexId);
                cellPanel.hide();
            } else {
                choiceParent = cell;
            }

            // add caption slots
            const top = html("div", { class: "info-text info-text-top" });
            const bot = html("div", { class: "info-text info-text-bottom" });
            cell.append(top, bot);

            function addToId(id) {
                CELLS.set(id, cell);

                SET_EVENT(`${id}/in`, () => cellPanel.in());
                SET_EVENT(`${id}/out`, () => cellPanel.out());
                
                SET_SPEAKER(`${id}/choice`, makeChoiceSpeaker(choiceParent));

                SET_SPEAKER(`${id}/caption`, async (...texts) => {
                    const parents = [top, bot];
        
                    for (let i = 0; i < texts.length; ++i) {
                        const text = texts[i];
                        const parent = parents[i % 2];
            
                        if (text.trim().length == 0) continue;
            
                        const voice = VOICES.get("caption");
                        const anim = new TextAnimation(text, voice, false);
                        parent.replaceChildren(html("span", {}, ...anim.glyphs.map((g) => g.element)));
                        await testAnim(anim);
                        await DELAY(1.5);
                        voice.clear();
                        await animateElement(parent.children[0], "caption-fadeout", .25);
                        anim.hide();
                    }
                });
            }

            if (panelData.id) {
                addToId(`${grid.id}/${panelData.id}`);
            }

            addToId(indexId);
        });

        logDatas.forEach((logData) => {
            const bounds = makePanel(logData);
            const cell = html("div", { style: "position: relative;", class: "log-cell" });
            animateElement(cell, "grow-out", 0);
            bounds.append(cell);

            copyStyle(cell, logData);

            const content = html("div", { class: "log-content text" });
            const panel = new Panel(content);
            const element = panel.root;
            element.classList.add("log-container");
            cell.append(element);
    
            const id = `${grid.id}/${logData.id}`;

            SET_EVENT(`${id}/in`, () => animateElement(cell, "grow-in", .25));
            SET_EVENT(`${id}/out`, () => animateElement(cell, "grow-out", .25));
    
            panel.addEventListener("pre-in", () => DELAY(0).then(() => scrollElementToBottom(content)));
    
            SET_SPEAKER(`${id}/say`, async (...texts) => {
                scrollElementToBottom(content);
    
                const anims = texts.map((text) => new TextAnimation(text, VOICES.get(logData.dataset.voice)));
                const pages = anims.map((anim) => ({ anim, root: html("div", {}, ...anim.glyphs.map((g) => g.element))}));
        
                const observer = new IntersectionObserver(offscreenCull);
                [...content.children].forEach((e) => observer.observe(e));
        
                for (const page of pages) {
                    const proceed = () => page.anim.proceed();
        
                    SIGNALS.addEventListener("advance", proceed);
                    
                    content.append(page.root);
                    scrollElementToBottom(content);
    
                    await testAnim(page.anim);
        
                    SIGNALS.removeEventListener("advance", proceed);
                    page.anim.finish();
    
                    observer.observe(page.root);
                }
    
                observer.disconnect();
            });
            SET_SPEAKER(`${id}/insert`, async (...texts) => {
                for (const text of texts) {
                    content.append(html("div", {}, text));
                }
                scrollElementToBottom(content);
            });
            SET_EVENT(`${id}/clear`, () => content.replaceChildren());
            SET_EVENT(`${id}/wipe`, async () => {
                const wipe = html("div", { class: "log-wipe" });
                element.append(wipe);
                await animateElement(wipe, "log-wipe", 1, "linear");
                content.replaceChildren();
                wipe.remove();
            });
        });

        speakerDatas.forEach((speakerData) => {
            const voice = speakerData.dataset.voice;
            const id = `${gridData.id}/${speakerData.id}`;

            const bounds = makePanel(speakerData);
            const cell = html("div", { style: "position: relative;", class: "speaker-cell" });
            bounds.append(cell);

            copyStyle(cell, speakerData);

            const options = {
                voice,
                portrait: speakerData.dataset.portrait,
                caption: speakerData.dataset.caption,
            }

            SET_SPEAKER(id, (...texts) => runCellSpeaker(cell, options, ...texts));
        });

        safe.append(grid);

        const children = [...grid.children];
        sortByIndex(children, (e) => e._childIndex);
        grid.replaceChildren(...children);
    });

    function setLoadingText(text) {
        document.getElementById("loading-container").replaceChildren(text);     
    }

    try {
        await PRELOAD_FONTS((progress) => setLoadingText(`loading fonts: ${Math.floor(progress*100)}%`));
        await PRELOAD_IMAGES((progress) => setLoadingText(`loading images: ${Math.floor(progress*100)}%`));
        await PRELOAD_MUSIC((progress) => setLoadingText(`loading music: ${Math.floor(progress*100)}%`));
        setLoadingText("tap or press key to begin");
    } catch (error) {
        setLoadingText("error loading. try refreshing");
        console.log(error);
        return;
    }

    await WAIT_AUDIO_ALLOWED();
    window.focus();
    document.getElementById("loading-container").remove();
    await RUN_EVENT("start");

    // run transcript
    while (commands.length > 0) {
        const command = commands.shift();

        if (command.type === "dialogue") {
            const { id, texts } = command;
            await RUN_SPEAKER(id, ...texts);
            await DELAY(.25);
        } else if (command.type === "event") {
            const { id, wait } = command;
            const event = RUN_EVENT(id);
            if (wait) await event;
        } else if (command.type === "javascript") {
            const { text } = command;
            const func = new AsyncFunction(text);
            await func();
        }
    }
}

test();
