/**
 * Get a child element matching CSS selector.
 * @param {string} query 
 * @param {ParentNode} element 
 * @returns {HTMLElement}
 */
 const ONE = (query, element = document) => element.querySelector(query);

 /**
  * Get all children elements matching CSS selector.
  * @param {string} query 
  * @param {ParentNode} element 
  * @returns {HTMLElement[]}
  */
 const ALL = (query, element = document) => Array.from(element.querySelectorAll(query));

/**
 * @template {keyof HTMLElementTagNameMap} K
 * @param {K} tagName 
 * @param {*} attributes 
 * @param  {...(Node | string)} children 
 * @returns {HTMLElementTagNameMap[K]}
 */
function html(tagName, attributes = {}, ...children) {
    const element = /** @type {HTMLElementTagNameMap[K]} */ (document.createElement(tagName));
    Object.entries(attributes).forEach(([name, value]) => element.setAttribute(name, value));
    children.forEach((child) => element.append(child));
    return element;
}

/** 
 * @param {number} milliseconds 
 * @returns {Promise}
 */
function sleep(milliseconds) {
    return new Promise(resolve => setTimeout(resolve, milliseconds));
}

function randFloat(min, max) {
    return min + (max - min) * Math.random();
}

/**
 * @param {HTMLElement}  element
 * @param {string} name
 * @param {number} seconds
 */
async function animateElement(element, name, seconds, ease="ease-in-out") {
    element.style.animation = `${name} ${seconds}s ${ease} forwards`;
    await sleep(seconds * 1000); 
}

// async equivalent of Function constructor
const AsyncFunction = Object.getPrototypeOf(async function () { }).constructor;

/**
 * @param {HTMLElement} element 
 * @param {boolean} integer
 */
function scaleElementToParent(element, integer = true) {
    const parent = element.parentElement;

    const [tw, th] = [parent.clientWidth, parent.clientHeight];
    const [sw, sh] = [tw / element.clientWidth, th / element.clientHeight];
    let scale = Math.min(sw, sh);
    scale = scale > 1 && integer ? Math.floor(scale) : scale;

    if (element.dataset.scale !== scale.toString()) {
        element.dataset.scale = scale.toString();
        element.style.setProperty("transform", `translate(-50%, -50%) scale(${scale})`);
    }

    return scale;
}


function cssTimeToMilliseconds(time_string) {
    var num = parseFloat(time_string),
        unit = time_string.match(/m?s/),
        milliseconds;

    if (unit) {
        unit = unit[0];
    }

    switch (unit) {
        case "s": // seconds
            milliseconds = num * 1000;
            break;
        case "ms": // milliseconds
            milliseconds = num;
            break;
        default:
            milliseconds = 0;
            break;
    }

    return milliseconds;
}

/**
 * @param {string} src
 * @returns {Promise<HTMLImageElement>}
 */
async function loadImage(src) {
    return new Promise((resolve, reject) => {
        const image = document.createElement("img");
        image.addEventListener("load", () => resolve(image));
        image.addEventListener("error", reject);
        image.src = src;
    });
}

const loadAudio = (src) => loadMediaElement(html("audio", { src }));
const loadVideo = (src) => loadMediaElement(html("video", { src }));

/**
 * @template {HTMLMediaElement} T 
 * @param {T} element
 * @returns {Promise<T>}
 */
async function loadMediaElement(element) {
    return new Promise((resolve, reject) => {
        let interval;
        
        function reset() {
            clearInterval(interval);
            interval = setInterval(() => resolve(element), 5000);
        }

        element.addEventListener("loadstart", reset);
        element.addEventListener("progress", reset);
        element.addEventListener("canplaythrough", () => resolve(element));
        element.addEventListener("error", () => reject(element));
        element.preload = "auto";
        element.load();
    });
}

/**
 * @param {HTMLElement} element
 */
function scrollElementToBottom(element) {
    element.scrollTop = element.scrollHeight;
}

function offscreenCull(entries) {
    entries.forEach(entry => {
        if (!entry.isIntersecting) entry.target.remove();
    });
}

/** 
 * @param {HTMLElement} target
 * @param {HTMLElement} source
 */
function copyStyle(target, source) {
    const classes = source.dataset.class?.split(" ") ?? [];
    target.classList.add(...classes);
    target.classList.add(...source.classList);
    target.style.cssText += `; ${source.style.cssText}`;
}

/** 
 * @template T
 * @param {T[]} array
 * @param {(item: T) => number} func
 */
function sortByIndex(array, func) {
    array.sort((a, b) => func(a) - func(b));
}
